from contextlib import contextmanager

import teradatasql
import time
import dbt_common.exceptions
import dbt.adapters.exceptions
from dbt.adapters.contracts.connection import AdapterResponse
from dbt.adapters.contracts.connection import Connection
from dbt.adapters.contracts.connection import Credentials
from dbt.adapters.events.logging import AdapterLogger
from dbt.adapters.sql import SQLConnectionManager
import re

logger = AdapterLogger("teradata")
from dataclasses import dataclass, field
from typing import Optional, Tuple, Any, Dict


@dataclass
class TeradataCredentials(Credentials):
    query_band: Optional[str] = field(default="org=teradata-internal-telem;appname=dbt;")                  # query_band parameter will be populated by profiles.yml, but it not a connection parameter so this won't be used to connect by driver
    server: Optional[str] = None
    database: Optional[str] = None
    schema: Optional[str] = None
    username: Optional[str] = None
    password: Optional[str] = None
    port: Optional[str] = None
    tmode: Optional[str] = "ANSI"
    logmech: Optional[str] = None
    charset: Optional[str] = None
    account: Optional[str] = None
    column_name: Optional[str] = None
    cop: Optional[str] = None
    coplast: Optional[str] = None
    encryptdata: Optional[str] = None
    fake_result_sets: Optional[str] = None
    field_quote: Optional[str] = None
    field_sep: Optional[str] = None
    lob_support: Optional[str] = None
    log: Optional[str] = None
    logdata: Optional[str] = None
    max_message_body: Optional[str] = None
    partition: Optional[str] = None
    sip_support: Optional[str] = None
    teradata_values: Optional[str] = None
    retries: int = 0
    retry_timeout: int = 1
    sslmode: Optional[str] = None
    sslca: Optional[str] = None
    sslcapath: Optional[str] = None
    sslcrc: Optional[str] = None
    sslcipher: Optional[str] = None
    sslprotocol: Optional[str] = None
    browser: Optional[str] = None
    browser_tab_timeout: Optional[int] = None
    browser_timeout: Optional[int] = None
    sp_spl: Optional[bool] = None
    sessions: Optional[int] = None
    runstartup: Optional[bool] = None
    logon_timeout: Optional[int] = None
    https_port: Optional[int] = None
    connect_timeout: Optional[int] = None
    request_timeout: Optional[int] = None
    http_proxy: Optional[str] = None
    http_proxy_user: Optional[str] = None
    http_proxy_password: Optional[str] = None
    https_proxy: Optional[str] = None
    https_proxy_user: Optional[str] = None
    https_proxy_password: Optional[str] = None
    sslcrl: Optional[bool] = None
    sslocsp: Optional[bool] = None
    proxy_bypass_hosts: Optional[str] = None
    oidc_sslmode: Optional[str] = None

    _ALIASES = {
        "UID": "username",
        "user": "username",
        "PWD": "password",
        "host": "server"
    }

    def __post_init__(self):
        if self.logmech is not None and self.logmech.lower() == "browser":
            # When logmech is "browser", username and password should not be provided.
            if self.username is not None or self.password is not None:
                raise dbt_common.exceptions.DbtRuntimeError(
                    "Couldn’t connect to Teradata Vantage SQL Engine. Neither username nor password parameters can be "
                    "specified in the profile when the logon mechanism (logmech) is ‘BROWSER'. Correct the profile "
                    "and retry.")
        else:
            if self.username is None:
                raise dbt_common.exceptions.DbtRuntimeError("Couldn’t  connect to Teradata Vantage SQL Engine. The "
                                                            "‘user’ parameter in the profile must be specified when "
                                                            "the logon mechanism (logmech) is ‘TD2'. Correct the "
                                                            "profile and retry.")
            elif self.password is None:
                raise dbt_common.exceptions.DbtRuntimeError("Couldn’t  connect to Teradata Vantage SQL Engine. The "
                                                            "‘password’ parameter in the profile must be specified "
                                                            "when the logon mechanism (logmech) is ‘TD2'. Correct the "
                                                            "profile and retry.")
        if self.schema is None:
            raise dbt_common.exceptions.DbtRuntimeError("Couldn’t  connect to Teradata Vantage SQL Engine. The "
                                                        "‘schema’ parameter in the profile must be specified . "
                                                        "Correct the profile and retry.")
        # teradata classifies database and schema as the same thing
        if (
                self.database is not None and
                self.database != self.schema
        ):
            raise dbt_common.exceptions.DbtRuntimeError(
                f"    schema: {self.schema} \n"
                f"    database: {self.database} \n"
                f"Couldn’t  connect to Teradata Vantage SQL Engine. The ‘database’ parameter in the profile is "
                f"specified and does not match the ‘schema’ parameter value. Correct the profile by removing the "
                f"‘database’ parameter or changing it to same value as ‘schema’ parameter and then retry."
            )
        if self.tmode == "TERA":
            note_for_tera = '''
----------------------------------------------------------------------------------
IMPORTANT NOTE: This is an initial implementation of the TERA transaction mode
and may not support some use cases.
We strongly advise validating all records or transformations utilizing this mode
to preempt any potential anomalies or errors
----------------------------------------------------------------------------------
            '''
            logger.info(note_for_tera)

    @property
    def type(self):
        return "teradata"

    @property
    def unique_field(self):
        return self.schema

    def _connection_keys(self):
        """
        Returns an iterator of keys to pretty-print in 'dbt debug'
        """
        return (
            "server",
            "port",
            "database",
            "schema",
            "user",
            "tmode",
            "logmech",
            "account",
            "column_name",
            "cop",
            "coplast",
            "encryptdata",
            "fake_result_sets",
            "field_quote",
            "field_sep",
            "lob_support",
            "log",
            "logdata",
            "max_message_body",
            "partition",
            "sip_support",
            "teradata_values",
            "sslmode",
            "sslca",
            "sslcapath",
            "sslcrc",
            "sslcipher",
            "sslprotocol",
            "browser",
            "browser_tab_timeout",
            "browser_timeout",
            "sp_spl",
            "sessions",
            "runstartup",
            "logon_timeout",
            "https_port",
            "connect_timeout",
            "request_timeout",
            "query_band",
            "http_proxy",
            "http_proxy_user",
            "http_proxy_password",
            "https_proxy",
            "https_proxy_user",
            "https_proxy_password",
            "sslcrl",
            "sslocsp",
            "proxy_bypass_hosts",
            "oidc_sslmode"
        )

    @classmethod
    def __pre_deserialize__(cls, data: Dict[Any, Any]) -> Dict[Any, Any]:
        # If database is not defined as adapter credentials
        data = super().__pre_deserialize__(data)
        if "database" not in data:
            data["database"] = None
        return data

class TeradataConnectionManager(SQLConnectionManager):
    TYPE = "teradata"

    '''
    disabling transactional logic by default for dbt-teradata
    by disabling add_begin_query(), add_commit_query(), begin(), commit()
    and clear_transaction() methods
    '''
    def add_begin_query(self):
        pass

    def add_commit_query(self):
        pass

    def begin(self):
        pass

    def commit(self):
        pass

    def clear_transaction(self):
        pass

    @classmethod
    def open(cls, connection):
        if connection.state == 'open':
            logger.debug('Connection is already open, skipping open.')
            return connection

        credentials = cls.get_credentials(connection.credentials)
        kwargs = {}

        kwargs["host"] = credentials.server
        kwargs["tmode"] = credentials.tmode
        if credentials.username:
           kwargs["user"] = credentials.username
        if credentials.password:
            kwargs["password"] = credentials.password
        if credentials.logmech:
            kwargs["logmech"] = credentials.logmech
        if credentials.account:
          kwargs["account"] = credentials.account
        if credentials.column_name:
          kwargs["column_name"] = credentials.column_name
        if credentials.cop:
          kwargs["cop"] = credentials.cop
        if credentials.coplast:
          kwargs["coplast"] = credentials.coplast
        if credentials.encryptdata:
          kwargs["encryptdata"] = credentials.encryptdata
        if credentials.fake_result_sets:
          kwargs["fake_result_sets"] = credentials.fake_result_sets
        if credentials.field_quote:
          kwargs["field_quote"] = credentials.field_quote
        if credentials.field_sep:
          kwargs["field_sep"] = credentials.field_sep
        if credentials.lob_support:
          kwargs["lob_support"] = credentials.lob_support
        if credentials.log:
          kwargs["log"] = credentials.log
        if credentials.logdata:
          kwargs["logdata"] = credentials.logdata
        if credentials.max_message_body:
          kwargs["max_message_body"] = credentials.max_message_body
        if credentials.partition:
          kwargs["partition"] = credentials.partition
        if credentials.sip_support:
          kwargs["sip_support"] = credentials.sip_support
        if credentials.teradata_values:
          kwargs["teradata_values"] = credentials.teradata_values
        if credentials.port:
            kwargs["dbs_port"] = credentials.port
        if credentials.sslmode:
           kwargs["sslmode"] = credentials.sslmode
        if credentials.sslca:
           kwargs["sslca"] = credentials.sslca
        if credentials.sslcapath:
           kwargs["sslcapath"] = credentials.sslcapath
        if credentials.sslcrc:
           kwargs["sslcrc"] = credentials.sslcrc
        if credentials.sslcipher:
           kwargs["sslcipher"] = credentials.sslcipher
        if credentials.sslprotocol:
           kwargs["sslprotocol"] = credentials.sslprotocol
        if credentials.browser:
           kwargs["browser"]=credentials.browser
        if credentials.browser_tab_timeout:
           kwargs["browser_tab_timeout"]=credentials.browser_tab_timeout
        if credentials.browser_timeout:
           kwargs["browser_timeout"]=credentials.browser_timeout
        if credentials.sp_spl:
           kwargs["sp_spl"]=credentials.sp_spl
        if credentials.sessions:
           kwargs["sessions"]=credentials.sessions
        if credentials.runstartup:
           kwargs["runstartup"]=credentials.runstartup
        if credentials.logon_timeout:
           kwargs["logon_timeout"]=credentials.logon_timeout
        if credentials.https_port:
           kwargs["https_port"]=credentials.https_port
        if credentials.connect_timeout:
           kwargs["connect_timeout"]=credentials.connect_timeout
        if credentials.request_timeout:
           kwargs["request_timeout"]=credentials.request_timeout
        if credentials.http_proxy:
            kwargs["http_proxy"]=credentials.http_proxy
        if credentials.http_proxy_user:
            kwargs["http_proxy_user"]=credentials.http_proxy_user
        if credentials.http_proxy_password:
            kwargs["http_proxy_password"]=credentials.http_proxy_password
        if credentials.https_proxy:
            kwargs["https_proxy"]=credentials.https_proxy
        if credentials.https_proxy_user:
            kwargs["https_proxy_user"]=credentials.https_proxy_user
        if credentials.https_proxy_password:
            kwargs["https_proxy_password"]=credentials.https_proxy_password
        if credentials.sslcrl:
            kwargs["sslcrl"]=credentials.sslcrl
        if credentials.sslocsp:
            kwargs["sslocsp"]=credentials.sslocsp
        if credentials.proxy_bypass_hosts:
            kwargs["proxy_bypass_hosts"]=credentials.proxy_bypass_hosts
        if credentials.oidc_sslmode:
            kwargs["oidc_sslmode"]=credentials.oidc_sslmode

        # Save the transaction mode
        cls.TMODE = credentials.tmode

        if credentials.retries > 0:
            def connect():
                connection.handle = teradatasql.connect(**kwargs)
                connection.state = 'open'

                # checking if query_band connection parameter exists. If it does -> executing the set query_band sql
                if credentials.query_band:
                    return cls.apply_query_band(connection.handle, credentials.query_band)

                return connection.handle

            retryable_exceptions = [teradatasql.OperationalError]

            return cls.retry_connection(
                connection,
                connect=connect,
                logger=logger,
                retry_limit=credentials.retries,
                retry_timeout=credentials.retry_timeout,
                retryable_exceptions=retryable_exceptions,
            )

        try:
            connection.handle = teradatasql.connect(**kwargs)
            connection.state = 'open'

            # checking if query_band connection parameter exists. If it does -> executing the set query_band sql
            if credentials.query_band:
                return cls.apply_query_band(connection.handle, credentials.query_band)

        except teradatasql.Error as e:
            logger.debug("Couldn’t  connect to Teradata Vantage SQL Engine. The Teradata driver error message is: '{}'"
                         "Correct the problem and retry."
                         .format(e))

            connection.handle = None
            connection.state = 'fail'

            raise dbt.adapters.exceptions.FailedToConnectError(str(e))

        return connection

    @classmethod
    def get_credentials(cls, credentials):
        return credentials

    def cancel(self, connection: Connection):
        connection.handle.close()

    @contextmanager
    def exception_handler(self, sql):
        try:
            yield

        except teradatasql.DatabaseError as e:
            logger.debug('Teradata error: {}'.format(str(e)))

            try:
                self.rollback_if_open()
            except teradatasql.Error:
                logger.debug("Failed to release connection!")
                pass

            raise dbt_common.exceptions.DbtDatabaseError(str(e).strip()) from e

        except Exception as e:
            logger.debug("Couldn’t  execute a SQL request against the Teradata Vantage SQL Engine. The SQL that "
                         "failed is: {}".format(sql))
            logger.debug("Rolling back transaction.")
            self.rollback_if_open()
            if isinstance(e, dbt_common.exceptions.DbtRuntimeError):
                # during a sql query, an internal to dbt exception was raised.
                # this sounds a lot like a signal handler and probably has
                # useful information, so raise it without modification.
                raise

            raise dbt_common.exceptions.DbtRuntimeError(e) from e

    @classmethod
    def get_response(cls, cursor):
        num_rows = 0
        activity = "success"
        message = "OK"
        if cursor is not None and cursor.rowcount is not None:
            num_rows = cursor.rowcount
            activity = cursor.activityname
            message = "activity: {}, rows_affected: {}".format(cursor.activityname, cursor.rowcount)
        return AdapterResponse(
          _message=message,
          rows_affected=num_rows,
          code=activity
        )
    '''
    overriding add_query method to disable
    transactional logic for dbt-teradata
    '''
    def add_query(
        self,
        sql: str,
        auto_begin: bool = False, #this avoid calling begin() method of SQLConnectionManager and hence disabling transactional logic
        bindings: Optional[Any] = None,
        abridge_sql_log: bool = False
    ) -> Tuple[Connection, Any]:
        try:
            return SQLConnectionManager.add_query(self, sql, auto_begin, bindings, abridge_sql_log)
        except Exception as ex:
            ignored = False
            query = sql.strip()
            if ("DROP view /*+ IF EXISTS */" in query) or ("DROP table /*+ IF EXISTS */" in query):
                for error_number in [3807, 3854, 3853]:
                    if f"[Error {error_number}]" in str (ex):
                        ignored = True
                        return None, None
            if ("DELETE DATABASE /*+ IF EXISTS */" in query) or ("DROP DATABASE /*+ IF EXISTS */" in query):
                for error_number in [3802]:
                    if f"[Error {error_number}]" in str (ex):
                        ignored = True
                        return None, None
            if not ignored:
                raise # rethrow

    # this method will return the datatype as string
    @classmethod
    def data_type_code_to_name(cls, type_code) -> str:
        return str(type_code)

    @classmethod
    def apply_query_band(cls, handle, query_band_text):
        try:
            cur = handle.cursor()

            if query_band_text[-1] != ';':
                query_band_text += ';'

            """checking if appname= key exist in query_band and if dbt exist as value of 
            that key otherwise appending +dbt in the end """
            if 'appname=' in query_band_text or 'appname =' in query_band_text:
                pair = query_band_text.split(';')
                for i in range(len(pair)):
                    if 'appname' in pair[i]:
                        val = pair[i].split('=')[1]
                        patterns = "^" + re.escape('dbt') + "$"
                        if not re.match(patterns, val):
                            pair[i] += '+dbt'
                query_band_text = ';'.join(pair)
            else:
                query_band_text = query_band_text + 'appname=dbt;'

            """ checking if 'org=' or 'org =' exist, if it doesn't appending 'org=teradata-internal-telem; in query_band 
                string.  If it exists, user might have set some value of their own, so doing nothing in that case"""

            if 'org=' not in query_band_text and 'org =' not in query_band_text:
                query_band_text += 'org=teradata-internal-telem;'

            query_band_str = "set query_band = '{}' for session;".format(query_band_text)
            cur.execute(query_band_str)
            cur.execute("sel GetQueryBand();")
            rows = cur.fetchone()
            logger.debug("Query Band set to {}".format(rows))           # To log in dbt.log
        except teradatasql.Error as ex:
            logger.debug(ex)
            logger.info("Couldn’t set the query_band using the specified value. Correct the query_band parameter in "
                        "the profile and retry.")
            raise dbt.exceptions.DbtRuntimeError(str(ex))

        return handle                                                     # returning the connection handle



from concurrent.futures import Future
from dataclasses import dataclass, asdict
from typing import Optional, List, Dict, Any, Union, Iterable, Callable, Set, FrozenSet, Tuple
import agate

import dbt
import dbt_common.exceptions

from dbt.adapters.base.impl import catch_as_completed, ConstraintSupport, AdapterConfig
from dbt_common.contracts.constraints import ConstraintType
from dbt.adapters.base.relation import InformationSchema
from dbt.adapters.sql import SQLAdapter
from dbt.adapters.teradata import TeradataConnectionManager
from dbt.adapters.teradata import TeradataRelation
from dbt.adapters.teradata import TeradataColumn
from dbt.adapters.capability import CapabilityDict, CapabilitySupport, Support, Capability
from dbt.adapters.base.meta import available
from dbt.adapters.base import BaseRelation
from dbt_common.clients.agate_helper import DEFAULT_TYPE_TESTER, table_from_rows
from dbt.adapters.contracts.relation import RelationType, RelationConfig
from dbt.adapters.events.logging import AdapterLogger
logger = AdapterLogger("teradata")
from dbt_common.utils import executor

LIST_SCHEMAS_MACRO_NAME = 'list_schemas'
LIST_RELATIONS_MACRO_NAME = 'list_relations_without_caching'
GET_CATALOG_MACRO_NAME = 'get_catalog'

def _expect_row_value(key: str, row: agate.Row):
    if key not in row.keys():
        raise dbt_common.exceptions.DbtInternalError(
            f'Got a row without \'{key}\' column, columns: {row.keys()}'
        )

    return row[key]

def _catalog_filter_schemas(used_schemas: FrozenSet[Tuple[str, str]]) -> Callable[[agate.Row], bool]:
    schemas = frozenset((None, s.lower()) for d, s in used_schemas)

    def test(row: agate.Row) -> bool:
        table_database = _expect_row_value('table_database', row)
        table_schema = _expect_row_value('table_schema', row)
        if table_schema is None:
            return False
        return (table_database, table_schema.lower()) in schemas

    return test

@dataclass
class TeradataConfig(AdapterConfig):
    query_band: Optional[str] = None

class TeradataAdapter(SQLAdapter):
    Relation = TeradataRelation
    Column = TeradataColumn
    ConnectionManager = TeradataConnectionManager

    CONSTRAINT_SUPPORT = {
        ConstraintType.check: ConstraintSupport.ENFORCED,
        ConstraintType.not_null: ConstraintSupport.ENFORCED,
        ConstraintType.unique: ConstraintSupport.ENFORCED,
        ConstraintType.primary_key: ConstraintSupport.ENFORCED,
        ConstraintType.foreign_key: ConstraintSupport.ENFORCED,
    }

    # adding full support for SchemaMetadataByRelations and TableLastModifiedMetadata in capability dictionary
    _capabilities: CapabilityDict = CapabilityDict(
    {
        Capability.SchemaMetadataByRelations: CapabilitySupport(support=Support.Full),
        Capability.TableLastModifiedMetadata: CapabilitySupport(support=Support.Full),
        Capability.MicrobatchConcurrency: CapabilitySupport(support=Support.Full)
    }
    )

    @classmethod
    def date_function(cls):
        return 'current_date()'

    @available
    def verify_database(self, database):
        if database.startswith('"'):
            database = database.strip('"')
        expected = self.config.credentials.schema
        if database.lower() != expected.lower():
            raise dbt_common.exceptions.NotImplementedError(
                'Cross-db references not allowed in {} ({} vs {})'
                .format(self.type(), database, expected)
            )
        # return an empty string on success so macros can call this
        return ''

    @classmethod
    def convert_text_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        return "LONG VARCHAR"

    @classmethod
    def convert_datetime_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        return "TIMESTAMP(0)"

    @classmethod
    def convert_time_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        return "TIME"

    @classmethod
    def convert_date_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        return "DATE"

    @classmethod
    def convert_boolean_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        return "BYTEINT"

    @classmethod
    def convert_number_type(
        cls, agate_table: agate.Table, col_idx: int
    ) -> str:
        decimals = agate_table.aggregate(agate.MaxPrecision(col_idx))
        return "FLOAT" if decimals else "INTEGER"

    @classmethod
    def quote(cls, identifier):
        return '"{}"'.format(identifier)

    def list_relations_without_caching(
        self, schema_relation: TeradataRelation
    ) -> List[TeradataRelation]:
        kwargs = {'schema_relation': schema_relation}
        try:
            results = self.execute_macro(
                LIST_RELATIONS_MACRO_NAME,
                kwargs=kwargs
            )
        except dbt_common.exceptions.DbtRuntimeError as e:
            errmsg = getattr(e, 'msg', '')
            if f"Teradata database '{schema_relation}' not found" in errmsg:
                return []
            else:
                logger.debug(f"Couldn’t retrieve relations from database {schema_relation}. An unexpected error was "
                             f"encountered. The Teradata error message is: {e.msg}. You may create a github issue "
                             f"with relevant details.")
                return []

        relations = []
        for row in results:
            if len(row) != 4:
                raise dbt_common.exceptions.DbtRuntimeError(
                    f"Invalid value from 'teradata__list_relations_without_caching({kwargs})', "
                    f"We expected 4 attributes but received {len(row)}. You may create a github issue with relevant "
                    f"details."
                )
            _, name, _schema, relation_type = row
            relation: BaseRelation = self.Relation.create(
                schema=_schema,
                identifier=name,
                type=relation_type
            )
            relations.append(relation)

        return relations

    def get_relation(
        self, database: str, schema: str, identifier: str
    ) -> Optional[BaseRelation]:
        if not self.Relation.get_default_include_policy().database:
            database = None

        return super().get_relation(database, schema, identifier)

    def get_catalog(
            self,
            relation_configs: Iterable[RelationConfig],
            used_schemas: FrozenSet[Tuple[str, str]],
        ) -> Tuple[agate.Table, List[Exception]]:
        schema_map = self._get_catalog_schemas(relation_configs)
        with executor(self.config) as tpe:
            futures: List[Future[agate.Table]] = []
            for info, schemas in schema_map.items():
                for schema in schemas:
                    futures.append(tpe.submit_connected(
                        self, schema,
                        self._get_one_catalog, info, [schema], used_schemas
                    ))
            catalogs, exceptions = catch_as_completed(futures)
        return catalogs, exceptions

    def _get_one_catalog(
        self,
        information_schema: InformationSchema,
        schemas: Set[str],
        relation_config: Iterable[RelationConfig],
    ) -> agate.Table:
        if len(schemas) != 1:
            raise dbt_common.exceptions.CompilationError(
                f'Couldn’t retrieve catalog. We expected 1 input schema in _get_one_catalog() but received {schemas}. '
                f'You may create a github issue with relevant details.'
            )

        return super()._get_one_catalog(information_schema, schemas, relation_config)

    @classmethod
    def _catalog_filter_table(cls, table: agate.Table, used_schemas: FrozenSet[Tuple[str, str]]) -> agate.Table:
        table = table_from_rows(
            table.rows,
            table.column_names,
            text_only_columns=['table_schema', 'table_name'],
        )
        return table.where(_catalog_filter_schemas(used_schemas))

    def check_schema_exists(self, database, schema):
        results = self.execute_macro(
            LIST_SCHEMAS_MACRO_NAME,
            kwargs={'database': database}
        )

        exists = True if schema in [row[0] for row in results] else False
        return exists

    # Methods used in adapter tests
    def update_column_sql(
        self,
        dst_name: str,
        dst_column: str,
        clause: str,
        where_clause: Optional[str] = None,
    ) -> str:
        clause = f'update {dst_name} set {dst_column} = {clause}'
        if where_clause is not None:
            clause += f' where {where_clause}'
        return clause

    def timestamp_add_sql(
        self, add_to: str, number: int = 1, interval: str = 'hour'
    ) -> str:
        # for backwards compatibility, we're compelled to set some sort of
        # default. A lot of searching has lead me to believe that the
        # '+ interval' syntax used in postgres/redshift is relatively common
        # and might even be the SQL standard's intention.
        return f"{add_to} + interval '{number}' {interval}"

    def string_add_sql(
        self, add_to: str, value: str, location='append',
    ) -> str:
        if location == 'append':
            return f"concat(cast(trim({add_to}) as varchar(63800)), '{value}')"
        elif location == 'prepend':
            return f"concat('{value}', cast(trim({add_to}) as varchar(63800))"
        else:
            raise dbt_common.exceptions.DbtRuntimeError(
                f'Couldn’t modify SQL string. The location parameter was neither "append" nor "prepend". Correct the '
                f'model and retry.'
            )

    def get_rows_different_sql(
        self,
        relation_a: TeradataRelation,
        relation_b: TeradataRelation,
        column_names: Optional[List[str]] = None,
    ) -> str:
        # This method only really exists for test reasons
        names: List[str]
        if column_names is None:
            columns = self.get_columns_in_relation(relation_a)
            names = sorted((self.quote(c.name) for c in columns))
        else:
            names = sorted((self.quote(n) for n in column_names))

        alias_a = "A"
        alias_b = "B"
        columns_csv_a = ', '.join([f"{alias_a}.{name}" for name in names])
        columns_csv_b = ', '.join([f"{alias_b}.{name}" for name in names])
        join_condition = ' AND '.join([f"{alias_a}.{name} = {alias_b}.{name}" for name in names])
        first_column = names[0]

        # Simulate an EXCEPT or MINUS operator
        COLUMNS_EQUAL_SQL = '''
        WITH
        a_except_b as (
            SELECT
                {columns_a}
            FROM {relation_a} as {alias_a}
            LEFT OUTER JOIN {relation_b} as {alias_b}
                ON {join_condition}
            WHERE {alias_b}.{first_column} is null
        ),
        b_except_a as (
            SELECT
                {columns_b}
            FROM {relation_b} as {alias_b}
            LEFT OUTER JOIN {relation_a} as {alias_a}
                ON {join_condition}
            WHERE {alias_a}.{first_column} is null
        ),
        diff_count as (
            SELECT
                1 as id,
                COUNT(*) as num_missing FROM (
                    SELECT * FROM a_except_b
                    UNION ALL
                    SELECT * FROM b_except_a
                ) as missing
        ),
        table_a as (
            SELECT COUNT(*) as num_rows FROM {relation_a}
        ),
        table_b as (
            SELECT COUNT(*) as num_rows FROM {relation_b}
        ),
        row_count_diff as (
            SELECT
                1 as id,
                table_a.num_rows - table_b.num_rows as difference
            FROM table_a, table_b
        )
        SELECT
            row_count_diff.difference as row_count_difference,
            diff_count.num_missing as num_mismatched
        FROM row_count_diff
        INNER JOIN diff_count ON row_count_diff.id = diff_count.id
        '''.strip()

        sql = COLUMNS_EQUAL_SQL.format(
            alias_a=alias_a,
            alias_b=alias_b,
            first_column=first_column,
            columns_a=columns_csv_a,
            columns_b=columns_csv_b,
            join_condition=join_condition,
            relation_a=str(relation_a),
            relation_b=str(relation_b),
        )

        return sql

    
    @available
    def standardize_grants_dict(self, grants_table: agate.Table) -> dict:
        """overridden this method as there were extra spaces that needed
         to be stripped
        """
        grants_dict: Dict[str, List[str]] = {}
        for row in grants_table:
            grantee = row["grantee"].strip()
            privilege = row["privilege_type"].strip()
            if privilege in grants_dict.keys():
                grants_dict[privilege].append(grantee)
            else:
                grants_dict.update({privilege: [grantee]})
        return grants_dict
    
    def valid_incremental_strategies(self):
        """The set of standard builtin strategies which this adapter supports out-of-the-box.
        Not used to validate custom strategies defined by end users.
        """
        return ["delete+insert","append","merge", "valid_history", "microbatch"]
    



from dataclasses import dataclass, field

from dbt.adapters.base.relation import BaseRelation, Policy
from dbt_common.exceptions import DbtRuntimeError

@dataclass
class TeradataQuotePolicy(Policy):
    database: bool = False
    schema: bool = True
    identifier: bool = True


@dataclass
class TeradataIncludePolicy(Policy):
    database: bool = False
    schema: bool = True
    identifier: bool = True


@dataclass(frozen=True, eq=False, repr=False)
class TeradataRelation(BaseRelation):
    quote_policy: Policy = field(default_factory=lambda: TeradataQuotePolicy())
    include_policy: Policy = field(default_factory=lambda: TeradataIncludePolicy())
    quote_character: str = '"'

    def render(self):
        if self.include_policy.database and self.include_policy.schema:
            raise DbtRuntimeError(
                f"Got a teradata relation with schema and database set to "
                "include, but only one can be set"
            )
        return super().render()

    ''' overriding render_limited() method because super method uses LIMIT clause which is not supported in Teradata
        This method is used when --empty flag in dbt run command is used for dry run of models '''
    def render_limited(self) -> str:
        rendered = self.render()
        if self.limit is None:
            return rendered
        elif self.limit == 0:
            return f"(select * from {rendered} sample 0) _dbt_limit_subq"
        else:
            return f"(select * from {rendered} sample {self.limit}) _dbt_limit_subq"



#!/usr/bin/env python
import os
import sys

from setuptools import setup


if sys.version_info < (3, 9) or sys.version_info >= (3, 14):
    print('Error: dbt-teradata does not support this version of Python.')
    print('Please install Python 3.9 or higher but less than 3.14.')
    sys.exit(1)


this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


package_name = "dbt-teradata"
package_version = "1.10.0"
description = """The Teradata adapter plugin for dbt (data build tool)"""


setup(
    name=package_name,
    version=package_version,

    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',

    author="Teradata Corporation",
    author_email="developers@teradata.com",
    url="https://github.com/Teradata/dbt-teradata",
    packages=[
        'dbt.adapters.teradata',
        'dbt.include.teradata',
    ],
    package_data={
        'dbt.include.teradata': [
            'macros/*.sql',
            'macros/materializations/**/*.sql',
            'macros/utils/*.sql',
            'macros/dbt_external_tables/helpers/*.sql',
            'macros/dbt_external_tables/*.sql',
            'dbt_project.yml',
            'sample_profiles.yml',
        ],
    },
    install_requires=[
        "dbt-adapters>=1.7.2",
        "dbt-common>=1.3.0",
        "teradatasql>=20.00.00.10",
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',

        'License :: OSI Approved :: Apache Software License',

        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',

        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
    ],
    python_requires=">=3.9,<3.14",
)
